open Syntax
open Printing
open EvalEnv
open EvalSubst


(************)
(* INTEGERS *)
(************)

(* Useful Constants *)
let zero = Constant (Int 0) 
let one = Constant (Int 1) 
let two = Constant (Int 2) 
let three = Constant (Int 3) 
let four = Constant (Int 4)

(*******************)
(* LET EXPRESSIONS *)
(*******************)

let let1 =
  Let ("x", zero,
  Let ("y", one,
  Op(Var "y", Plus, Var "x")))

let let2 =
  Let ("x", zero,
  Let ("x", one,
  Op(Var "x", Plus, Var "x")))

let let3 =
  Let ("x", one,
  Let ("y", (Let ("x", four, Op(Var "x",Minus,one))),
  Op(Var "x", Times, Var "y")))

(***********************)
(* RECURSIVE FUNCTIONS *)
(***********************)

(***********************)
(* RECURSIVE FUNCTIONS *)
(***********************)

(* let z = 2 in (let x = 3 in fun y -> x + y + z) (let x = 4 in x + z) *)
let clo =  
  Let ("z", two,
       App (Let ("x", three, 
		 Rec ("f", "y", 
		      Op (Var "x", Plus, Op (Var "y", Plus, Var "z")))
                ),
	    Let ("x", four, Op (Var "x", Plus, Var "z"))
           )
      )

(* rec fact n = if n < 1 then 1 else n * fact (n - 1) *)
let fact = 
  Rec ("fact", "n", 
       If (Op (Var "n", Less, one),
           one,
           Op (Var "n", Times, 
               App (Var "fact", 
                        Op (Var "n", Minus, one)))))

(* fact 4 *)
let fact4 = App (fact, four)

(*********)
(* PAIRS *)
(*********)

(* the pair (1,2) *)
let p1 = Pair (one, two)

(* the function swap below is equivalent to:
       let swap p = let (x,y) = p in (y,x)
*)
let swap = 
  Rec ("swap", "p",
       Let ("x", Fst (Var "p"),
       Let ("y", Snd (Var "p"),
       Pair(Var "y", Var "x"))))

(* use swap to swap the elements of p1 *)
let swap_p1 = App (swap, p1)

(*********)
(* LISTS *)
(*********)

(* takes an OCaml list of expressions and generates a single expression
   representing the list
*)
let rec listify (l:exp list) : exp =
  match l with
      [] -> EmptyList
    | hd::tl -> Cons(hd,listify tl)

(* a list of 4 numbers *)
let list4 = listify [one;two;three;four] 

(* rec sumlist l = 
 *   match l with
 *     [] -> 0
 *   | hd::tl -> hd + sumlist tl *)
let sumlist = 
  Rec ("sumlist", "l", 
       Match (Var "l",
           zero,
           "hd", "tl", Op (Var "hd", Plus, 
			   App (Var "sumlist", Var "tl"))))

let sl4 = App (sumlist, list4)

(*****************)
(* EXAMPLE TESTS *)
(*****************)

(* STUDENTS:  EACH EXAMPLE IN THE LIST BELOW
 * WILL BE EXECUTED WITH BOTH SUBSTITUTION-BASED
 * AND ENVIRONMENT-BASED INTERPRETER.
 * ADD AND REMOVE EXAMPLES TO TEST WHAT YOU'VE DONE!
 *)
let examples = [
  zero;    (* 0 *)
  one;     (* 1 *)
  p1;      (* (1,2) *)
  let1;
  let2;
  let3;
  swap_p1; (* swap (1,2) *)
  fact4;   (* fact 4 *)
  sl4;     (* sumlist [1;2;3;4] *)
]

let example_num = List.length examples

(*****************************************)
(* Property-based Testing Infrastructure *)
(*****************************************)

(* specifies a property-based test *)
type 'a tester = 
  {property : 'a -> bool; 
   error    : 'a -> unit; 
   generate : unit -> 'a;}

let test_many (msg:string) (max:int) (tester:'a tester): unit =
  assert (max >= 0);
  let rec loop (i:int) (successes:int): int =
    if i < max then (
      let data = tester.generate () in
      if tester.property data then (
        loop (i+1) 
      (successes + 1) 
      ) else (
        tester.error data; 
        loop (i+1) successes
      )
    ) else (
      successes
    )
  in
  Printf.printf "Testing %s:\n" msg;
  let successes = loop 0 0 in
  Printf.printf "\n%d out of %d successful tests\n" successes max

let nl () = print_newline()

(*
 * We cannot reasonably generate "random" expressions that typecheck. Therefore 
 * we will need to "generate" them from a predetermined list. This list will be
 * a list of expressions that are the conversions of strings imported from files.
 * To be extra precise, this is the workflow to create this list:
 * 
 * NOTE: things you have to do are marked (Y)
 *
 *  - Create a new file in the `Tests` folder (Y).
 *  - Write an expression (as a string) that uses only the "MiniMl" subset (Y).
 *  - Add this filename (including the `Tests/` prefix) in the list `example_files` (Y).
 *  - For each string in `example_files`, open this file and convert its contents to an expression.
 * 
 * Done, you have a list of expressions.
 *)

(************************************************)
(* property-based testing of EvalEnv: *)
(************************************************)

(* mutable ref points to example expression list; no need to understand this *)
let next_expr = ref examples

let p1_generate () : Syntax.exp = 
  match (!next_expr) with
  | [] -> failwith "Generated more expressions than provided."
  | hd::tl -> next_expr := tl; hd

let p1 (e: Syntax.exp) : bool =
  let eval_subst =
    try Some (EvalSubst.eval e)
    with
    | _ -> (
      Printf.printf 
        "%s\n%s\n" 
        ("EvalSubst raises exception: "^(Printing.string_of_exp e))
        ("Successful test count will be off by 1");
      None)
  in
  let eval_env =
    try Some (EvalEnv.eval e)
    with
    | _ -> (
      Printf.printf 
        "%s\n%s\n" 
        ("EvalEnv raises exception: "^(Printing.string_of_exp e))
        ("Successful test count will be off by 1"); 
      None)
  in
  match eval_subst, eval_env with
  | Some es, Some ee -> es = ee
  | _, _ -> true  (* continue testing if an exception was raised 
                     this will give bad count... *)

let p1_error (e: Syntax.exp) : unit =
  Printf.printf "EVALUATION OF EvalEnv AND EvalSubst DO NOT MATCH:\n";
  let eval_subst = EvalSubst.eval e in
  let eval_env = EvalEnv.eval e in
  Printf.printf "Input:"; print_string (Printing.string_of_exp e); nl();
  Printf.printf "EvalSubst output:"; print_string (Printing.string_of_exp eval_subst);
  Printf.printf "EvalEnv output:"; print_string (Printing.string_of_exp eval_env);
  nl()

let tester1 = {property=p1; error = p1_error; generate=p1_generate;}

(********)
(* Main *)
(********)

let main () =
  nl();
  test_many "EvalEnv VS EvalSubst" example_num tester1 ; nl(); (* this line tests evaluation of EvalEnv and EvalSubst *)
  Printf.printf "DONE"; nl()

let _ = main ()